"""
Kubernetes POD EXEC transport
=============================

Performs `exec` operation into EXISTING, RUNNING POD to run a backup operation in-place.
"""

import time
from typing import List
from kubernetes import config, client
from kubernetes.client import V1PodList, V1Pod, V1ObjectMeta
from rkd.api.inputoutput import IO

from bahub.bin import RequiredBinary, copy_required_tools_from_controller_cache_to_target_env, \
    copy_encryption_keys_from_controller_to_target_env
from bahub.settings import BIN_VERSION_CACHE_PATH, TARGET_ENV_BIN_PATH, TARGET_ENV_VERSIONS_PATH
from bahub.transports.base import TransportInterface, create_backup_maker_command
from bahub.transports.kubernetes import KubernetesPodFilesystem, pod_exec, ExecResult
from bahub.transports.sh import LocalFilesystem


class Transport(TransportInterface):
    _v1_core_api: client.CoreV1Api
    _v1_apps_api: client.AppsV1Api
    _process: ExecResult
    _binaries: List[RequiredBinary]

    _namespace: str
    _selector: str
    _io: IO

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)
        self._namespace = spec.get('namespace')
        self._selector = spec.get('selector')

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "oneOf": [
                {"required": ["namespace", "selector"]},
            ],
            "properties": {
                "selector": {
                    "type": "string",
                    "example": "my-label=myvalue",
                    "default": ""
                },
                "namespace": {
                    "type": "string",
                    "example": "prod",
                    "default": "default"
                }
            }
        }

    @property
    def v1_core_api(self) -> client.CoreV1Api:
        if not hasattr(self, '_v1_core_api'):
            config.load_kube_config()  # todo: Add support for selecting cluster
            self._v1_core_api = client.CoreV1Api()

        return self._v1_core_api

    @property
    def v1_apps_api(self) -> client.AppsV1Api:
        if not hasattr(self, '_v1_apps_api'):
            config.load_kube_config()  # todo: Add support for selecting cluster
            self._v1_apps_api = client.AppsV1Api()

        return self._v1_apps_api

    def prepare_environment(self, binaries: List[RequiredBinary]) -> None:
        self._binaries = binaries

    def schedule(self, command: str, definition, is_backup: bool, version: str = "") -> None:
        """
        Runs a `kubectl exec` on already existing POD
        """

        pod_name = self.find_pod_name(self._selector, self._namespace)
        self._execute_in_pod_when_pod_will_be_ready(pod_name, command, definition, is_backup, version)

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        """
        Todo: Delete GPG keys from POD

        :param exc_type:
        :param exc_val:
        :param exc_t:
        :return:
        """
        pass

    def _execute_in_pod_when_pod_will_be_ready(self, pod_name: str, command: str, definition,
                                               is_backup: bool, version: str = ""):
        """
        Spawns backup process in a prepared environment inside POD
        Waits for POD to be ready, injects required dependencies then starts a command
        Later command will be watched using watch() API method

        :param pod_name:
        :param command:
        :param definition:
        :param is_backup:
        :param version:
        :return:
        """

        self.wait_for_pod_to_be_ready(pod_name, self._namespace)
        self._prepare_environment_inside_pod(definition, pod_name)

        complete_cmd = create_backup_maker_command(command, definition, is_backup, version,
                                                   bin_path=TARGET_ENV_BIN_PATH)
        self.io().debug(f"POD exec: `{complete_cmd}`")

        self._process = pod_exec(
            pod_name=pod_name,
            namespace=self._namespace,
            cmd=complete_cmd,
            io=self._io
        )

    def _prepare_environment_inside_pod(self, definition, pod_name: str) -> None:
        """
        Populate with GPG keys and required tools

        :param definition:
        :param pod_name:
        :return:
        """

        pod_fs = KubernetesPodFilesystem(pod_name, self._namespace, self.io())
        copy_encryption_keys_from_controller_to_target_env(
            src_fs=LocalFilesystem(),
            pub_key_path=definition.encryption().get_public_key_path(),
            private_key_path=definition.encryption().get_private_key_path(),
            dst_fs=pod_fs,
            io=self.io()
        )
        copy_required_tools_from_controller_cache_to_target_env(
            local_cache_fs=LocalFilesystem(),
            dst_fs=pod_fs,
            io=self.io(),
            bin_path=TARGET_ENV_BIN_PATH,
            versions_path=TARGET_ENV_VERSIONS_PATH,
            local_versions_path=BIN_VERSION_CACHE_PATH,
            binaries=self._binaries
        )

    def watch(self) -> bool:
        """
        Buffers stdout/stderr to io.debug() and notifies about exit code at the end
        """

        self._process.watch(self.io().debug)
        return self._process.has_exited_with_success()

    def wait_for_pod_to_be_ready(self, pod_name: str, namespace: str, timeout: int = 120):
        """
        Waits for POD to reach a valid state

        :raises: When timeout hits
        """

        self.io().debug("Waiting for POD to be ready...")

        for i in range(0, timeout):
            pod: V1Pod = self._v1_core_api.read_namespaced_pod(name=pod_name, namespace=namespace)

            if pod.status.phase in ["Ready", "Healthy", "True", "Running"]:
                return True

            self.io().debug(f"Pod not ready. Status: {pod.status.phase}")
            time.sleep(1)

        raise Exception(f"Timed out while waiting for pod '{pod_name}' in namespace '{namespace}'")

    def find_pod_name(self, selector: str, namespace: str):
        """
        Returns a POD name

        :raises: When no matching POD found
        """

        pods: V1PodList = self.v1_core_api.list_namespaced_pod(namespace,
                                                               label_selector=selector,
                                                               limit=1)

        if len(pods.items) == 0:
            raise Exception(f'No pods found matching selector {selector} in {namespace} namespace')

        pod: V1Pod = pods.items[0]
        pod_metadata: V1ObjectMeta = pod.metadata

        self.io().debug(f"Found POD name: '{pod_metadata.name}' in namespace '{namespace}'")

        return pod_metadata.name

    def get_required_binaries(self):
        return []



import json
import yaml
from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext, TaskInterface
from ..importing import Importing
from ..transports.all import transports
from ..adapters import adapters


class BackupTypeSchemaPrintingTask(TaskInterface):
    """Shows JSON schema of section 'spec' in backup of given type"""

    def get_name(self) -> str: return ':schema'
    def get_group_name(self) -> str: return ':help:backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        parser.add_argument('type', help='Task type (value of the field "type" in backup definitions)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        task_type = context.get_arg('type')
        adapter, definition = Importing.import_adapter(task_type)

        self.io().outln(
            json.dumps(definition.get_specification_schema(), indent=4, sort_keys=True)
        )

        return True


class BackupTypeExampleTask(TaskInterface):
    """Shows example configuration for a given backup type. See :help:info for list of built-in tasks"""

    def get_name(self) -> str: return ':example'
    def get_group_name(self) -> str: return ':help:backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        parser.add_argument('type', help='Task type (value of the field "type" in backup definitions)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        task_type = context.get_arg('type')
        adapter, definition = Importing.import_adapter(task_type)

        self.io().outln(
            yaml.dump({
                'backups': {
                    'example': definition.get_example_configuration()
                }
            }, indent=4)
        )

        return True


class TransportTypeTask(TaskInterface):
    """Shows example transport configuration for given transport. See :help:info for list of built-in transports"""

    def get_name(self) -> str: return ':example'
    def get_group_name(self) -> str: return ':help:transport'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = False):
        parser.add_argument('type', help='Transport type (value of the field "type" in transport list)')

    def execute(self, context: ExecutionContext) -> bool:
        super().execute(context)

        transport_name = context.get_arg('type')
        transport = Importing.import_transport(transport_name)

        self.io().outln(
            yaml.dump({
                'transports': {
                    'example': transport.get_example_configuration()
                }
            }, indent=4)
        )

        return True


class InfoTask(TaskInterface):
    """Lists all built-in backup types and transports"""

    def get_name(self) -> str: return ':info'
    def get_group_name(self) -> str: return ':help'

    def execute(self, context: ExecutionContext):

        self.io().print_line()
        self.io().outln('Standard built-in transports:')

        for transport in transports():
            self.io().outln('- {}'.format(transport.__module__))

        self.io().print_line()
        self.io().outln('Standard built-in backup types:')

        for adapter in adapters():
            self.io().outln('- {}'.format(adapter.__module__))

        self.io().print_line()
        self.io().info_msg('*The list includes ONLY officially distributed adapters and transports. '
                           'You may be able to install additional adapters and transports via PyPI')

        self.io().info_msg('Check also help for transports: `bahub :help:transport:example bahub.transports.sh`')
        self.io().info_msg('Check also help for backup types: `bahub :help:backup:example bahub.adapters.filesystem`')

        return True

    def configure_argparse(self, parser: ArgumentParser):
        pass




from abc import ABC as AbstractClass, abstractmethod


class Response(AbstractClass):
    @abstractmethod
    def to_status_message(self) -> str:
        """Formats the response as user-readable to show user as a message"""

        pass




import os
from unittest.mock import patch
from rkd.api.inputoutput import IO, BufferedSystemIO
from rkd.api.testing import BasicTestingCase
from bahub.testing import create_example_fs_definition, run_transport
from bahub.transports.sh import Transport


class TestShellTransport(BasicTestingCase):
    """
    Functional test - requires /bin/bash
    """

    @patch('bahub.transports.sh.create_backup_maker_command')
    def test_executes_command_locally_and_it_returns_fine(self, create_backup_maker_command):
        transport = self._create_example_transport(BufferedSystemIO())
        definition = create_example_fs_definition(transport)

        # our current working directory should be reachable by the transport
        create_backup_maker_command.return_value = ["test", "-d", os.getcwd()]

        self.assertTrue(run_transport(definition, transport))

    @patch('bahub.transports.sh.create_backup_maker_command')
    def test_executes_command_locally_and_returns_failure_when_command_fails(self, create_backup_maker_command):
        transport = self._create_example_transport(BufferedSystemIO())
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["/bin/false"]

        self.assertFalse(run_transport(definition, transport))

    @patch('bahub.transports.sh.create_backup_maker_command')
    def test_fails_when_command_does_not_exists(self, create_backup_maker_command):
        io = BufferedSystemIO()
        transport = self._create_example_transport(io)
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["some-not-existing-command"]

        self.assertFalse(run_transport(definition, transport))
        self.assertIn("No such file or directory", io.get_value())

    @staticmethod
    def _create_example_transport(io: IO) -> Transport:
        return Transport(
            spec={'shell': '/bin/bash'},
            io=io
        )



from argparse import ArgumentParser
from rkd.api.contract import ExecutionContext
from .base import BaseTask


class RestoreTask(BaseTask):
    """Restores data from previously uploaded backup
    """

    def get_name(self) -> str:
        return ':restore'

    def get_group_name(self) -> str:
        return ':backup'

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        super().configure_argparse(parser, with_definition=with_definition)
        parser.add_argument('--version', default='latest', help='Version number. Defaults to "latest"')
        parser.add_argument('--download', '-d', required=False, help='Set target path as argument to download only')

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        return self.call_backup_maker(context, is_backup=True, version=context.get_arg('--version'))




"""
Shell Transport
===============

Executes a command in the shell
"""
import os
import shutil
import subprocess
import sys
from tempfile import TemporaryDirectory
from typing import List

from rkd.api.inputoutput import IO

from .base import TransportInterface, create_backup_maker_command
from ..bin import RequiredBinary, download_required_tools, copy_required_tools_from_controller_cache_to_target_env
from ..fs import FilesystemInterface
from ..inputoutput import StreamableBuffer
from ..model import BackupDefinition


class LocalFilesystem(FilesystemInterface):
    io: IO

    def force_mkdir(self, path: str):
        try:
            os.mkdir(path)
        except FileExistsError:
            pass

    def download(self, url: str, destination_path: str):
        subprocess.check_call(["curl", "-s", "-L", "--output", destination_path, url])

    def delete_file(self, path: str):
        os.unlink(path)

    def link(self, src: str, dst: str):
        os.link(src, dst)

    def make_executable(self, path: str):
        subprocess.check_call(["chmod", "+x", path])

    def file_exists(self, path: str) -> bool:
        return os.path.isfile(path)

    def pack(self, archive_path: str, src_path: str, files_list: List[str]):
        if not files_list:
            files_list = ["*", ".*"]

        subprocess.check_call(["tar", "-zcf", archive_path] + files_list, cwd=src_path)

    def copy_to(self, local_path: str, dst_path: str):
        shutil.copyfile(local_path, dst_path)

    def unpack(self, archive_path: str, dst_path: str):
        subprocess.check_call(["tar", "-xf", archive_path, "--directory", dst_path])

    def find_temporary_dir_path(self) -> str:
        return TemporaryDirectory().name

    def move(self, src: str, dst: str):
        subprocess.check_call(["mv", src, dst])


class Transport(TransportInterface):
    """
    Local shell transport
    =====================

    Allows to execute commands in a local shell
    """

    handle: StreamableBuffer
    bin_path: str = os.path.expanduser("~/.backuprepository/bin")
    versions_path: str = os.path.expanduser("~/.backuprepository/bin/versions")
    fs: FilesystemInterface

    error: Exception

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)
        self.fs = LocalFilesystem()

    def prepare_environment(self, binaries: List[RequiredBinary]) -> None:
        copy_required_tools_from_controller_cache_to_target_env(
            local_cache_fs=LocalFilesystem(),
            dst_fs=self.fs,
            io=self.io(),
            bin_path=self.bin_path,
            versions_path=self.versions_path,
            local_versions_path=os.path.expanduser("~/.backup-controller/versions"),
            binaries=binaries
        )

    def schedule(self, command: str, definition: BackupDefinition, is_backup: bool, version: str = "") -> None:
        try:
            self.handle = self._exec_command(
                create_backup_maker_command(command, definition, is_backup, version), env={
                    "PATH": os.getenv("PATH") + ":" + self.bin_path
                }
            )
        except FileNotFoundError as exc:
            self.error = exc

    def watch(self) -> bool:
        if hasattr(self, "error") and self.error:
            self.io().error(str(self.error))
            return False

        self.handle.copy_to(sys.stderr)
        self.handle.close()

        return self.handle.finished_with_success()

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": [],
            "properties": {

            }
        }



from typing import Union

from rkd.api.inputoutput import BufferedSystemIO
from rkd.api.testing import BasicTestingCase
from bahub.bin import RequiredBinary, RequiredBinaryFromGithubRelease, RequiredBinaryFromGithubReleasePackedInArchive, \
    download_required_tools, copy_encryption_keys_from_controller_to_target_env, \
    copy_required_tools_from_controller_cache_to_target_env
from bahub.fs import FilesystemInterface


class TestRequiredBinary(BasicTestingCase):
    def test_get_full_name_with_version(self):
        self.assertEqual("vunknown-kubectl",
                         RequiredBinary("https://example.org/releases/kubectl").get_full_name_with_version())

    def test_is_archive(self):
        self.assertTrue(RequiredBinary("https://example.org/releases/kubectl.tar.gz").is_archive())

        # other types than tar.gz are not supported
        self.assertFalse(RequiredBinary("https://example.org/releases/kubectl.zip").is_archive())
        self.assertFalse(RequiredBinary("https://example.org/releases/kubectl").is_archive())

    def test_get_filename(self):
        self.assertEqual("kubectl", RequiredBinary("https://example.org/releases/kubectl").get_filename())


class TestRequiredBinaryFromGithubRelease(BasicTestingCase):
    def test_get_url(self):
        binary = RequiredBinaryFromGithubRelease("riotkit-org/tracexit", "1.0.0", "tracexit")

        self.assertEqual("https://github.com/riotkit-org/tracexit/releases/download/1.0.0/tracexit", binary.get_url())


class TestRequiredBinaryFromGithubReleasePackedInArchive(BasicTestingCase):
    def test_functional(self):
        binary = RequiredBinaryFromGithubReleasePackedInArchive(
            project_name="riotkit-org/tracexit",
            version="1.0.0",
            binary_name="tracexit",
            archive_name="tracexit-1.0.0-amd64.tar.gz"
        )

        self.assertTrue(binary.is_archive())
        self.assertEqual("tracexit", binary.get_filename())
        self.assertEqual(
            "https://github.com/riotkit-org/tracexit/releases/download/1.0.0/tracexit-1.0.0-amd64.tar.gz",
            binary.get_url()
        )


class FSMock(object):
    callstack: list

    def __init__(self):
        self.callstack = []

    def __getattr__(self, item):
        return lambda *args, **kwargs: self.callstack.append([item, args, kwargs])


class TestDownloadRequiredTools(BasicTestingCase):
    """
    Covers download_required_tools()
    """

    def test_downloads_and_unpacks_archive(self):
        """
        When tool does not exist, then it should be downloaded and unpacked (when it is an archive)

        :return:
        """

        io = BufferedSystemIO()
        io.set_log_level("debug")

        fs: Union[FilesystemInterface, FSMock] = FSMock()
        fs.file_exists = lambda path: False
        fs.find_temporary_dir_path = lambda: "/tmp/test"

        download_required_tools(
            fs=fs,
            io=io,
            bin_path="/opt/bin",
            versions_path="/opt/bin/.versions",
            binaries=[
                RequiredBinaryFromGithubReleasePackedInArchive(
                    project_name="riotkit-org/tracexit",
                    version="1.0.0",
                    binary_name="tracexit",
                    archive_name="tracexit-1.0.0-amd64.tar.gz"
                )
            ]
        )

        # directory structure is created
        self.assertIn(['force_mkdir', ('/opt',), {}], fs.callstack)
        self.assertIn(['force_mkdir', ('/opt/bin',), {}], fs.callstack)
        self.assertIn(['force_mkdir', ('/opt/bin/.versions',), {}], fs.callstack)

        # binary is downloaded into temporary directory
        self.assertIn(
            ['download', ('https://github.com/riotkit-org/tracexit/releases/download/1.0.0/tracexit-1.0.0-amd64.tar.gz',
                          '/tmp/test/archive.tar.gz'), {}],
            fs.callstack
        )

        # archive is unpacked
        self.assertIn(
            ['unpack', ('/tmp/test/archive.tar.gz', '/tmp/test'), {}],
            fs.callstack
        )

        # unpacked binary file is moved to .versions directory
        self.assertIn(
            ['move', ('/tmp/test/tracexit', '/opt/bin/.versions/v1.0.0-tracexit'), {}],
            fs.callstack
        )

        # file is made executable, so it can be executed later
        self.assertIn(
            ['make_executable', ('/opt/bin/.versions/v1.0.0-tracexit',), {}],
            fs.callstack
        )

    def test_file_is_not_downloaded_twice(self):
        io = BufferedSystemIO()
        io.set_log_level("debug")

        fs: Union[FilesystemInterface, FSMock] = FSMock()
        fs.file_exists = lambda path: True  # file ALREADY EXISTS

        download_required_tools(
            fs=fs,
            io=io,
            bin_path="/opt/bin",
            versions_path="/opt/bin/.versions",
            binaries=[
                RequiredBinaryFromGithubReleasePackedInArchive(
                    project_name="riotkit-org/tracexit",
                    version="1.0.0",
                    binary_name="tracexit",
                    archive_name="tracexit-1.0.0-amd64.tar.gz"
                )
            ]
        )

        self.assertNotIn("download", str(fs.callstack))

    def test_file_is_downloaded_and_not_unpacked_if_not_an_archive(self):
        io = BufferedSystemIO()
        io.set_log_level("debug")

        fs: Union[FilesystemInterface, FSMock] = FSMock()
        fs.file_exists = lambda path: False
        fs.find_temporary_dir_path = lambda: "/tmp/test"

        download_required_tools(
            fs=fs,
            io=io,
            bin_path="/opt/bin",
            versions_path="/opt/bin/.versions",
            binaries=[
                RequiredBinary("https://bakunin.org/binary-name")
            ]
        )

        self.assertIn(
            ['make_executable', ('/opt/bin/.versions/vunknown-binary-name',), {}],
            fs.callstack
        )

        # should be downloaded
        self.assertIn("download", str(fs.callstack))

        # but should not be unpacked, as not an archive
        self.assertNotIn("unpack", str(fs.callstack))


class TestEncryptionKeysCopying(BasicTestingCase):
    """
    Covers copy_encryption_keys_from_controller_to_target_env()
    """

    def test_copies_only_keys_that_exists(self):
        io = BufferedSystemIO()
        io.set_log_level("debug")

        src_fs: Union[FilesystemInterface, FSMock] = FSMock()

        # mock: Only public key exists. Private key does not exist
        src_fs.file_exists = lambda path: path == "id_rsa.pub"

        dst_fs: Union[FilesystemInterface, FSMock] = FSMock()

        copy_encryption_keys_from_controller_to_target_env(
            src_fs=src_fs,
            dst_fs=dst_fs,
            pub_key_path="id_rsa.pub",
            private_key_path="id_rsa",
            io=io
        )

        self.assertIn(
            ['copy_to', ('id_rsa.pub', '/tmp/.gpg.pub'), {}],
            dst_fs.callstack
        )

        # private key is not copied, as it does not exist
        self.assertNotIn(
            ['copy_to', ('id_rsa', '/tmp/.gpg.key'), {}],
            dst_fs.callstack
        )


class TestToolsTransferFromLocalCacheToTargetFilesystem(BasicTestingCase):
    """
    Covers copy_required_tools_from_controller_cache_to_target_env()
    """

    def test_no_binaries_triggers_early_exit(self):
        fs: Union[FilesystemInterface, FSMock] = FSMock()
        io = BufferedSystemIO()
        io.set_log_level("debug")

        copy_required_tools_from_controller_cache_to_target_env(
            local_cache_fs=fs, dst_fs=fs, bin_path="/opt/bin", versions_path="/opt/bin/.versions",
            local_versions_path="/tmp/.versions",
            binaries=[],  # this is empty,
            io=io
        )

        self.assertIn("All binaries are up-to-date", io.get_value())

    def test_archive_is_copied_and_tools_are_linked(self):
        local_cache_fs: Union[FilesystemInterface, FSMock] = FSMock()
        dst_fs: Union[FilesystemInterface, FSMock] = FSMock()
        io = BufferedSystemIO()
        io.set_log_level("debug")

        copy_required_tools_from_controller_cache_to_target_env(
            local_cache_fs=local_cache_fs, dst_fs=dst_fs, bin_path="/opt/bin", versions_path="/opt/bin/.versions",
            local_versions_path="/tmp/.versions",
            binaries=[
                RequiredBinaryFromGithubReleasePackedInArchive(
                    project_name="riotkit-org/tracexit",
                    version="1.6.1",
                    binary_name="tracexit",
                    archive_name="tracexit-1.0.0-amd64.tar.gz"
                ),
                RequiredBinaryFromGithubReleasePackedInArchive(
                    project_name="riotkit-org/br-backup-maker",
                    version="2.1.3.7",
                    binary_name="br-backup-maker",
                    archive_name="br-backup-maker-1.0.0-amd64.tar.gz"
                )
            ],
            io=io
        )

        # versions are packed at source filesystem
        self.assertIn('pack', str(local_cache_fs.callstack))
        self.assertIn("'v1.6.1-tracexit', 'v2.1.3.7-br-backup-maker'", str(local_cache_fs.callstack))

        # versions are unpacked at target filesystem
        self.assertIn("copy_to", str(dst_fs.callstack))
        self.assertIn(
            ['unpack', ('/tmp/.backup-tools.tar.gz', '/opt/bin/.versions'), {}],
            dst_fs.callstack
        )

        # files are linked
        self.assertIn(
            ['link', ('/opt/bin/.versions/v1.6.1-tracexit', '/opt/bin/tracexit'), {}],
            dst_fs.callstack
        )
        self.assertIn(
            ['link', ('/opt/bin/.versions/v2.1.3.7-br-backup-maker', '/opt/bin/br-backup-maker'), {}],
            dst_fs.callstack
        )



import os

HOME_PATH = os.path.expanduser("~/.backup-controller")
BIN_CACHE_PATH = HOME_PATH + "/bin"
BIN_VERSION_CACHE_PATH = BIN_CACHE_PATH + '/versions'
CONFIG_PATH = os.path.expanduser("~/.backup-controller/config.yaml")

TARGET_ENV_BIN_PATH = "/tmp/.br"
TARGET_ENV_VERSIONS_PATH = "/tmp/.br/versions"



import os
from abc import ABC
from argparse import ArgumentParser
from traceback import print_exc
from typing import Dict, Union, Optional, List
from rkd.api.contract import TaskInterface, ExecutionContext, ArgumentEnv
from rkd.yaml_parser import YamlFileLoader
from rkd.exception import YAMLFileValidationError

from ..adapters.base import AdapterInterface
from ..api import BackupRepository
from ..bin import get_backup_maker_binaries, download_required_tools, RequiredBinary
from ..configurationfactory import ConfigurationFactory
from ..model import BackupDefinition
from ..notifier import MultiplexedNotifiers, NotifierInterface
from ..security import create_sensitive_data_stripping_filter
from ..settings import BIN_CACHE_PATH, BIN_VERSION_CACHE_PATH, CONFIG_PATH
from ..transports.sh import LocalFilesystem


class BaseTask(TaskInterface, ABC):
    config: ConfigurationFactory
    api: BackupRepository
    notifier: Union[MultiplexedNotifiers, NotifierInterface]

    def get_declared_envs(self) -> Dict[str, Union[str, ArgumentEnv]]:
        return {
            'CONFIG': ArgumentEnv('CONFIG', '--config', CONFIG_PATH),
            'DEBUG': ArgumentEnv('DEBUG', '--debug', '')
        }

    def execute(self, context: ExecutionContext) -> bool:
        try:
            self.config = ConfigurationFactory(
                configuration_path=context.get_arg_or_env('--config'),
                debug=bool(context.get_arg_or_env('--debug')),
                parser=YamlFileLoader([]),
                io=self._io
            )

        except YAMLFileValidationError as e:
            self.io().error('Configuration file looks invalid, details: ' + str(e))
            print_exc()
            return False

        if not context.get_arg('--show-secrets'):
            self._io.add_output_processor(create_sensitive_data_stripping_filter(self.config.get_all_sensitive_data()))

        self.api = BackupRepository(self._io)
        self.notifier = MultiplexedNotifiers(self.config.notifiers())

        return True

    def prepare_binaries_cache(self, binaries: List[RequiredBinary]):
        """
        Binaries cache keeps single-binary applications required to perform a backup
        Files are copied from the cache to the destination environment e.g. Docker container or Kubernetes POD
        """

        download_required_tools(
            fs=LocalFilesystem(),
            io=self.io(),
            bin_path=BIN_CACHE_PATH,
            versions_path=BIN_VERSION_CACHE_PATH,
            binaries=binaries
        )

    def call_backup_maker(self, context: ExecutionContext, is_backup: bool, version: str = "") -> bool:
        """
        Schedules a "Backup Maker" to perform action.

        The scheduling is done by using a Transport layer which takes responsibility for allocating resources,
        spawning the process and tracking it.
        """

        definition_name = context.get_arg('definition')
        definition = self.config.get_definition(definition_name)
        adapter: AdapterInterface = self.config.get_adapter(definition_name)()
        required_binaries = adapter.get_required_binaries() + get_backup_maker_binaries() + \
                            definition.get_transport_required_tools()

        self.prepare_binaries_cache(required_binaries)

        # begin a backup, get a buffered reader
        with definition.transport(binaries=required_binaries) as transport:
            self.notifier.starting_backup_creation(definition)

            if is_backup:
                transport.schedule(adapter.create_backup_instruction(definition), definition,
                                   is_backup=True, version=version)
            else:
                transport.schedule(adapter.create_restore_instruction(definition), definition,
                                   is_backup=False, version=version)

            self.io().info("Listening to backup-maker logs (through transport)")
            is_success = transport.watch()
            self.io().info("backup-maker process finished")

        if not is_success:
            self.io().error_msg('Process did not return success. Check previous messages for details')

            additional_info = transport.get_failure_details()

            if additional_info:
                self.io().outln(additional_info)

        return is_success

    def configure_argparse(self, parser: ArgumentParser, with_definition: bool = True):
        # do not require as switch, allow to use env
        parser.add_argument('--config', '-c', default=CONFIG_PATH, required=False)
        parser.add_argument('--debug', action='store_true')
        parser.add_argument('--show-secrets', action='store_true', help='Do not hide secrets in output')

        if with_definition:
            parser.add_argument('definition', help='Backup definition name from the configuration file')

    def retrieve_definition(self, context: ExecutionContext) -> Optional[BackupDefinition]:
        definition_name = context.get_arg('definition')
        definition = self.config.get_definition(definition_name)

        if not definition:
            self.io().error('Definition "{name}" is not valid'.format(name=definition_name))

        return definition



"""
Core Tasks
==========

Provides shell commands such as:
- preparing backups
- sending backups
- listing backups
- receiving backups

"""
from rkd.api.syntax import TaskDeclaration
from .prepare import BackupPreparationTask
from .restore import RestoreTask
from .docs import BackupTypeSchemaPrintingTask, BackupTypeExampleTask, TransportTypeTask, InfoTask


def imports():
    return [
        TaskDeclaration(RestoreTask()),
        TaskDeclaration(BackupPreparationTask()),
        TaskDeclaration(BackupTypeSchemaPrintingTask()),
        TaskDeclaration(BackupTypeExampleTask()),
        TaskDeclaration(TransportTypeTask()),
        TaskDeclaration(InfoTask())
    ]



"""
Server API connector
====================

Notice: Only E2E tests coverage
"""
import certifi
import pycurl
from subprocess import Popen, PIPE
from io import BytesIO
from json import JSONDecodeError, loads as json_loads
from rkd.api.inputoutput import IO

from bahub.env import is_curl_debug_mode
from bahub.exception import InvalidResponseException, HttpException
from bahub.inputoutput import StreamableBuffer
from bahub.model import ServerAccess


class BackupRepository(object):
    def __init__(self, io: IO):
        self.io = io

    # todo: implement
    def validate_self_permissions(self):
        pass

    # todo: implement
    def create_temporary_token(self):
        pass

    # # todo: remove
    # def send_backup(self, collection_id: str, access: ServerAccess, attributes: VersionAttributes,
    #                 source: StreamableBuffer) -> VersionUploadedResponse:
    #
    #     response_body_stream = BytesIO()
    #     url = access.build_url(
    #         endpoint=SEND_BACKUPS_URL.replace('{collectionId}', collection_id),
    #         attributes=attributes
    #     )
    #
    #     curl = pycurl.Curl()
    #     curl.setopt(pycurl.URL, url)
    #     curl.setopt(pycurl.CAINFO, certifi.where())
    #     curl.setopt(pycurl.CUSTOMREQUEST, 'POST')
    #     curl.setopt(pycurl.UPLOAD, 1)
    #     curl.setopt(pycurl.READFUNCTION, source.read)
    #     curl.setopt(pycurl.WRITEFUNCTION, response_body_stream.write)
    #     curl.setopt(pycurl.HTTPHEADER, ["Authorization: Bearer " + access.get_token()])
    #
    #     # debug
    #     if is_curl_debug_mode():
    #         curl.setopt(pycurl.VERBOSE, True)
    #         curl.setopt(pycurl.SSL_VERIFYPEER, 0)
    #         curl.setopt(pycurl.SSL_VERIFYHOST, 0)
    #
    #     # explicit timeouts
    #     curl.setopt(pycurl.CONNECTTIMEOUT, 300)
    #     curl.setopt(pycurl.TIMEOUT, 3600*24)
    #
    #     try:
    #         curl.perform()
    #
    #     except pycurl.error as e:
    #         self.io.error('Curl error cause: {}'.format(e.__cause__))
    #
    #         raise HttpException(
    #             'HTTP request error: ' + str(e) + '. Probably the application ' +
    #             'backup process exited or timed out unexpectedly. Read above messages for details')
    #
    #     response_body = response_body_stream.getvalue().decode('utf-8')
    #
    #     self.io.debug('Request: ' + str(url))
    #     self.io.debug('response_body=' + response_body)
    #
    #     try:
    #         _json = json_loads(response_body)
    #     except JSONDecodeError:
    #         _json = {}
    #
    #     response_code = curl.getinfo(pycurl.HTTP_CODE)
    #
    #     self.io.debug('response_code={}'.format(response_code))
    #
    #     if response_code >= 400:
    #         raise InvalidResponseException('HTTP error, code {}, body: {}'.format(response_code, response_body), _json,
    #                                        _json.get('error_code', 0))
    #
    #     curl.close()
    #
    #     return VersionUploadedResponse(
    #         version=_json['version']['version'],
    #         fileid=_json['version']['id'],
    #         filename=_json['version']['file']['filename']
    #     )
    #
    # # todo: remove
    # def read_backup(self, collection_id: str, version: str, access: ServerAccess) -> StreamableBuffer:
    #     url = access.build_url(
    #         RECEIVE_BACKUPS_URL.format(collectionId=collection_id, version=version)
    #     )
    #
    #     process = ['curl', '--silent', '-X', 'GET', '-H', 'Authorization: Bearer {}'.format(access.get_token()), url]
    #
    #     self.io.debug('read_backup({})'.format(process))
    #
    #     # cannot be done with requests or urllib3 due to raw IO stream handles required to copy streams
    #     # from process to process. Requests uses urllib3, which produces invalid streams that produces deadlocks
    #     proc = Popen(process, stdout=PIPE)
    #
    #     return StreamableBuffer(
    #         io=self.io,
    #         read_callback=proc.stdout.read,
    #         close_callback=lambda: proc.stdout.close(),
    #         eof_callback=lambda: proc.poll() is not None,
    #         is_success_callback=lambda: proc.poll() == 0,
    #         has_exited_with_failure=lambda: proc.poll() != 0,
    #         description='API file read stream<{}, {}>'.format(collection_id, version),
    #         buffer=proc.stdout
    #     )



"""
Docker Transport
================

Executes commands inside a running docker container.

Note: Requires access to the docker daemon. Make sure your user is in a "docker" group (have access to the socket)
"""
import os
import subprocess
from tempfile import TemporaryDirectory
import docker
from typing import List, Generator
from docker import DockerClient
from docker.models.containers import Container
from rkd.api.inputoutput import IO
from .base import TransportInterface, create_backup_maker_command
from .sh import LocalFilesystem
from ..bin import RequiredBinary, copy_required_tools_from_controller_cache_to_target_env
from ..fs import FilesystemInterface
from ..settings import TARGET_ENV_BIN_PATH, TARGET_ENV_VERSIONS_PATH


class DockerFilesystemTransport(FilesystemInterface):
    container: Container
    io: IO

    def __init__(self, container: Container):
        self.container = container

    def force_mkdir(self, path: str):
        exit_code, result = self.container.exec_run(["mkdir", "-p", path])
        assert exit_code == 0, f"Cannot create directory inside container at path {path}: {result}"

    def download(self, url: str, destination_path: str):
        exit_code, result = self.container.exec_run(["curl", "-f", "-L", "--output", destination_path, url])
        assert exit_code == 0, f"Cannot download {url} into {destination_path} inside container: {result}"

    def delete_file(self, path: str):
        exit_code, result = self.container.exec_run(["rm", "-f", path])
        assert exit_code == 0, f"Cannot remove file at path {path} inside container: {result}"

    def link(self, src: str, dst: str):
        exit_code, result = self.container.exec_run(["ln", "-s", src, dst])
        assert exit_code == 0, f"Cannot link {src} to {dst} inside container: {result}"

    def make_executable(self, path: str):
        exit_code, result = self.container.exec_run(["chmod", "+x", path])
        assert exit_code == 0, f"Cannot make file executable at path '{path}' inside container: {result}"

    def file_exists(self, path: str):
        exit_code, result = self.container.exec_run(["test", "-f", path])
        return exit_code == 0

    def copy_to(self, local_path: str, dst_path: str):
        subprocess.check_call(["docker", "cp", local_path, self.container.id + ":" + dst_path])

    def pack(self, archive_path: str, src_path: str, files_list: List[str]):
        if not files_list:
            files_list = ["*", ".*"]

        exit_code, result = self.container.exec_run(["tar", "-zcf", archive_path] + files_list, workdir=src_path)
        assert exit_code == 0, f"Cannot pack '{src_path}'/* into {archive_path} (both paths inside container). {result}"

    def unpack(self, archive_path: str, dst_path: str):
        exit_code, result = self.container.exec_run(["tar", "-xf", archive_path, "--directory", dst_path])
        assert exit_code == 0, f"Cannot unpack tar archive from '{archive_path}' to '{dst_path}' " \
                               f"(both paths inside container). {result}"

    def find_temporary_dir_path(self) -> str:
        return TemporaryDirectory().name

    def move(self, src: str, dst: str):
        exit_code, result = self.container.exec_run(["mv", src, dst])
        return exit_code == 0, f"Cannot move file '{src}' to '{dst}'"


class Transport(TransportInterface):
    """
    Docker Transport
    ================

    Enables a hot-backup inside a running application container
    """

    _container_name: str
    _shell: str
    _client: DockerClient
    _container: Container
    bin_path: str = TARGET_ENV_BIN_PATH
    versions_path: str = TARGET_ENV_VERSIONS_PATH
    binaries: List[RequiredBinary]
    _exec_stream: Generator
    _exec_id: str

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)

        self._spec = spec
        self._io = io
        self._container_name = spec.get('container')
        self._shell = spec.get('shell', '/bin/sh')

    @property
    def client(self) -> DockerClient:
        if not hasattr(self, "_client"):
            self._client = docker.from_env()
            self._populate_container_information()

        return self._client

    def _populate_container_information(self):
        self.fs = DockerFilesystemTransport(self.container)

    @property
    def container(self) -> Container:
        if not hasattr(self, "_container"):
            self._container = self.client.containers.get(self._container_name)

        return self._container

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["container"],
            "properties": {
                "container": {
                    "type": "string",
                    "example": "my_docker_container_name_1"
                },
                "shell": {
                    "type": "string",
                    "example": "/bin/sh",
                    "default": "/bin/sh"
                }
            }
        }

    def prepare_environment(self, binaries: List[RequiredBinary]) -> None:
        self.binaries = binaries

    def schedule(self, command: str, definition, is_backup: bool, version: str = "") -> None:
        """
        Runs a command inside a container

        :param command:
        :param definition:
        :param is_backup:
        :param version:
        :return:
        """

        new_path = self.discover_path_variable_in_container() + ":" + self.bin_path
        self.io().debug(f"Setting $PATH={new_path}")

        copy_required_tools_from_controller_cache_to_target_env(
            local_cache_fs=LocalFilesystem(),
            dst_fs=self.fs,
            io=self.io(),
            bin_path=self.bin_path,
            versions_path=self.versions_path,
            local_versions_path=os.path.expanduser("~/.backup-controller/versions"),  # todo
            binaries=self.binaries
        )

        complete_cmd = create_backup_maker_command(command, definition, is_backup, version)

        self.io().debug(f"Docker exec: {complete_cmd}")

        # spawn command
        response = self.client.api.exec_create(
            self.container.id,
            complete_cmd,
            environment={
                'PATH': new_path
            }
        )

        # start spawned command. Save its ID - we will be able to track its status later
        self._exec_id = response['Id']
        self._exec_stream = self.client.api.exec_start(
            response['Id'], stream=True
        )

    def watch(self) -> bool:
        """
        Watches live stream from `docker exec` and asserts exit_code == 0
        :return:
        """

        for lines in self._exec_stream:
            for line in lines.split(b'\n'):
                self.io().info(line.decode('utf-8'))

        self.io().debug(f"Docker exec process returned code={self._exit_code}")
        return self._exit_code == 0

    def discover_path_variable_in_container(self) -> str:
        """
        Returns $PATH value from container
        :return:
        """

        exit_code, output = self.container.exec_run(['/bin/sh', '-c', 'echo "$PATH"'])

        return str(output.decode('utf-8')).strip()

    @property
    def _exit_code(self) -> int:
        return int(self.client.api.exec_inspect(self._exec_id)['ExitCode'])

    def get_failure_details(self) -> str:
        return "Error occurred while trying to execute command inside docker container - {}, \n" \
               "=======> LOGS FROM CONTAINER:\n {} =======> END OF LOGS"\
            .format(
                self._container_name,
                self._try_to_collect_logs()
            )

    def _try_to_collect_logs(self) -> str:
        """
        Try to collect logs (if it is possible) from the container
        """

        try:
            return self.container.logs(tail=15, follow=False, stream=False).decode('utf-8')

        except Exception:
            return '-- No logs --'



import os

import pytest
from testcontainers.core.container import DockerContainer
from unittest.mock import patch
from rkd.api.inputoutput import IO, BufferedSystemIO
from rkd.api.testing import BasicTestingCase
from bahub.transports.docker_sidecontainer import Transport as SideDockerTransport
from bahub.transports.docker import Transport as DockerExecTransport
from bahub.testing import create_example_fs_definition


class TestDockerTransport(BasicTestingCase):
    """
    Functional test - requires docker daemon and docker client tools
    """

    _nginx_container: DockerContainer

    @patch('bahub.transports.docker.create_backup_maker_command')
    def test_side_docker_transport_executes_command_inside_temporary_container(self, create_backup_maker_command):
        io = BufferedSystemIO()
        transport = self._create_example_side_docker_transport(io)
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["ls", "-la", "/mnt"]

        self.assertTrue(self._run_transport(definition, transport), msg='Expected that `ls -la /mnt` would not fail')

        # at /mnt was mounted ./.github directory from repository (see setUp())
        self.assertIn('workflows', io.get_value(),
                      msg="Failed asserting that backup container on Alpine Linux has same volumes "
                          "as application container with NGINX image")

    @pytest.mark.xfail  # randomly fails on CI
    @patch('bahub.transports.docker.create_backup_maker_command')
    def test_docker_exec_transport_executes_command_inside_container(self, create_backup_maker_command):
        io = BufferedSystemIO()
        transport = self._create_example_docker_exec_transport(io)
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["ls", "/config"]

        self.assertTrue(self._run_transport(definition, transport), msg="Expected that `ls /config` would work")

        # assert that according to https://github.com/linuxserver/docker-nginx/pkgs/container/nginx
        # there are a few directories, so the `docker exec` operation works
        self.assertIn("www", io.get_value())
        self.assertIn("php", io.get_value())
        self.assertIn("nginx", io.get_value())

    @patch('bahub.transports.docker.create_backup_maker_command')
    def test_docker_exec_transport_reports_failure_when_binary_not_found(self, create_backup_maker_command):
        io = BufferedSystemIO()
        transport = self._create_example_docker_exec_transport(io)
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["not-a-valid-command"]

        self.assertFalse(self._run_transport(definition, transport),
                         msg="Expected that invalid command will result in a failure")

        self.assertIn("executable file not found in $PATH", io.get_value(),
                      msg="Expected that there will be any message")

    @patch('bahub.transports.docker.create_backup_maker_command')
    def test_docker_exec_transport_reports_failure_when_command_fails(self, create_backup_maker_command):
        io = BufferedSystemIO()
        transport = self._create_example_docker_exec_transport(io)
        definition = create_example_fs_definition(transport)

        create_backup_maker_command.return_value = ["/bin/sh", "-c", "/bin/false"]

        self.assertFalse(self._run_transport(definition, transport), msg='/bin/false should return exit code 1')

    @staticmethod
    def _run_transport(definition, transport) -> bool:
        with definition.transport(binaries=[]):
            transport.schedule(
                command="--mocked--", definition=definition,
                is_backup=True, version=""
            )

            return transport.watch()

    # ================
    # Technical stuff
    # ================

    def setUp(self) -> None:
        super().setUp()
        self._nginx_container = DockerContainer(image='ghcr.io/linuxserver/nginx:1.20.2')\
            .with_name("nginx-app").with_volume_mapping(os.getcwd() + "/.github", "/mnt").start()

    def tearDown(self) -> None:
        super().tearDown()
        self._nginx_container.stop(force=True, delete_volume=True)

    @staticmethod
    def _create_example_side_docker_transport(io: IO) -> SideDockerTransport:
        return SideDockerTransport(
            spec={
                'orig_container': "nginx-app",
                'temp_container_image': 'ghcr.io/mirrorshub/docker/alpine:3.14',
                'shell': '/bin/sh'
            },
            io=io
        )

    @staticmethod
    def _create_example_docker_exec_transport(io: IO) -> DockerExecTransport:
        return DockerExecTransport(
            spec={
                'container': "nginx-app",
                'shell': '/bin/sh'
            },
            io=io
        )



"""
Backup Binaries
===============

Single-binary helpers are used to perform more sophisticated operations e.g. perform database queries,
gracefully shutdown a system, block incoming traffic etc.

Those binaries are fetched to the local cache, then are distributed to remote filesystems where the backup is performed.
"""


import os.path
from typing import List
from tempfile import NamedTemporaryFile
from rkd.api.inputoutput import IO
from bahub.fs import FilesystemInterface
from bahub.versions import BACKUP_MAKER_BIN_VERSION, TRACEXIT_BIN_VERSION


class RequiredBinary(object):
    """
    Binary file downloadable from specified URL address
    """

    url: str

    def __init__(self, url: str):
        self.url = url

    def get_version(self) -> str:
        return "unknown"

    def get_filename(self):
        return os.path.basename(self.url)

    def get_full_name_with_version(self) -> str:
        return f"v{self.get_version()}-{self.get_filename()}"

    def get_url(self):
        return self.url

    def is_archive(self) -> bool:
        return self.url.endswith('tar.gz')


class RequiredBinaryFromGithubRelease(RequiredBinary):
    """
    Binary file released on GitHub
    """

    version: str
    binary_name: str

    def __init__(self, project_name: str, version: str, binary_name: str):
        self.version = version
        self.binary_name = binary_name

        super().__init__("https://github.com/{project_name}/releases/download/{version}/{binary_name}".format(
            project_name=project_name, version=version, binary_name=binary_name
        ))

    def get_version(self) -> str:
        return self.version

    def get_filename(self) -> str:
        return self.binary_name


class RequiredBinaryFromGithubReleasePackedInArchive(RequiredBinaryFromGithubRelease):
    """
    Binary file released on GitHub as tar.gz packaged archive
    (e.g. by GoReleaser)
    """

    def __init__(self, project_name: str, version: str, binary_name: str, archive_name: str):
        super().__init__(project_name, version, archive_name)
        self.binary_name = binary_name

    def is_archive(self) -> bool:
        return True


def download_required_tools(fs: FilesystemInterface, io: IO, bin_path: str,
                            versions_path: str, binaries: List[RequiredBinary]) -> None:
    """
    Collects all binaries VERSIONED into /bin/versions
    Does not download binary twice.

    Actually this method should be used to download tools into local `Backup Controller` cache at first stage.
    At later stage - tools are copied to target environment and symbolic links are used.
    """

    io.debug("Preparing environment")
    fs.force_mkdir(os.path.dirname(bin_path))
    fs.force_mkdir(bin_path)
    fs.force_mkdir(versions_path)

    for binary in binaries:
        version_path = versions_path + "/" + binary.get_full_name_with_version()

        io.debug(f"Searching for tool {version_path}")

        if not fs.file_exists(version_path):
            io.debug(f"Downloading tool {binary.get_url()} into {version_path}")

            if binary.is_archive():
                tmp_dir = fs.find_temporary_dir_path()
                fs.force_mkdir(tmp_dir)
                fs.download(binary.get_url(), tmp_dir + "/archive.tar.gz")
                fs.unpack(tmp_dir + "/archive.tar.gz", tmp_dir)
                fs.move(tmp_dir + "/" + binary.get_filename(), version_path)
                fs.make_executable(version_path)
            else:
                fs.download(binary.get_url(), version_path)
                fs.make_executable(version_path)


def copy_encryption_keys_from_controller_to_target_env(src_fs: FilesystemInterface, dst_fs: FilesystemInterface,
                                                       pub_key_path: str, private_key_path: str, io: IO) -> None:
    """
    Copies GPG keys from source to REMOTE filesystem

    :param src_fs:
    :param dst_fs:
    :param pub_key_path:
    :param private_key_path:
    :param io:
    :return:
    """

    io.info("Copying encryption keys")

    for key_type, key_path in {'key': private_key_path, 'pub': pub_key_path}.items():
        if key_path and src_fs.file_exists(key_path):
            io.debug(f"{key_path} -> /tmp/.gpg.{key_type}")
            dst_fs.copy_to(key_path, f"/tmp/.gpg.{key_type}")


def copy_required_tools_from_controller_cache_to_target_env(local_cache_fs: FilesystemInterface,
                                                            dst_fs: FilesystemInterface, io: IO,
                                                            bin_path: str, versions_path: str, local_versions_path: str,
                                                            binaries: List[RequiredBinary]):
    """
    Pack selected binaries from local cache, send them to remote filesystem and unpack

    :param local_cache_fs: Local filesystem where we store cache
    :param dst_fs: Destination filesystem e.g. Kubernetes POD's FS or docker container FS
    :param io:
    :param bin_path: dst_fs's part of $PATH (where symbolic links are stored)
    :param versions_path: dst_fs's path where the versioned binaries are stored
    :param local_versions_path:
    :param binaries:
    :return:
    """

    io.info("Copying required tools from scheduler to Backup Maker target environment")
    selected_files_to_transfer = []

    # 1: Collect list of binaries that needs to be packed into archive
    for binary in binaries:
        version_path = versions_path + "/" + binary.get_full_name_with_version()

        if not dst_fs.file_exists(version_path):
            selected_files_to_transfer.append(binary.get_full_name_with_version())

    io.info(f"Missing binaries: {selected_files_to_transfer}. Will be copied to target environment")

    if not selected_files_to_transfer:
        io.info(f"All binaries are up-to-date")
        return

    # 2: Pack everything into archive
    with NamedTemporaryFile() as tmp_archive_path:
        local_cache_fs.pack(tmp_archive_path.name, local_versions_path, selected_files_to_transfer)

        # 3: Unpack archive at destination filesystem
        io.debug(f"Unpacking at {versions_path}")
        dst_fs.copy_to(tmp_archive_path.name, '/tmp/.backup-tools.tar.gz')
        dst_fs.force_mkdir(bin_path)
        dst_fs.force_mkdir(versions_path)
        dst_fs.unpack('/tmp/.backup-tools.tar.gz', versions_path)

    # 3: Link versioned files into generic names e.g. "v1.2.3-pg-backuper" into "pg-backuper"
    for binary in binaries:
        target_bin_path = bin_path + "/" + binary.get_filename()
        version_path = versions_path + "/" + binary.get_full_name_with_version()

        io.debug(f"Linking version {version_path} into {target_bin_path}")
        dst_fs.delete_file(target_bin_path)
        dst_fs.link(version_path, target_bin_path)
        dst_fs.make_executable(version_path)


def get_backup_maker_binaries() -> List[RequiredBinary]:
    return [
        RequiredBinaryFromGithubReleasePackedInArchive(
            project_name="riotkit-org/br-backup-maker",
            version=BACKUP_MAKER_BIN_VERSION,
            binary_name="br-backup-maker",
            # todo: support for multiple architectures
            archive_name="br-backup-maker_0.0.4_linux_amd64.tar.gz"
        ),
        RequiredBinaryFromGithubReleasePackedInArchive(
            project_name="riotkit-org/tracexit",
            version=TRACEXIT_BIN_VERSION,
            binary_name="tracexit",
            # todo: support for multiple architectures
            archive_name=f"tracexit_{TRACEXIT_BIN_VERSION}_linux_amd64.tar.gz"
        )
    ]



"""
Transport
=========

Abstraction that allows end-user to select if the backup should be done via local shell, ssh, docker,
kubernetes container etc.

Any transport can be implemented that will take a command, then process and properly output the result.
"""



from abc import abstractmethod
from typing import List


class FilesystemInterface(object):
    """
    Interacts with filesystem
    Implementations should allow interacting with filesystems in various places like remote filesystems or containers
    """

    @abstractmethod
    def force_mkdir(self, path: str):
        pass

    @abstractmethod
    def download(self, url: str, destination_path: str):
        pass

    @abstractmethod
    def delete_file(self, path: str):
        pass

    @abstractmethod
    def link(self, src: str, dst: str):
        pass

    @abstractmethod
    def make_executable(self, path: str):
        pass

    @abstractmethod
    def copy_to(self, local_path: str, dst_path: str):
        pass

    @abstractmethod
    def pack(self, archive_path: str, src_path: str, files_list: List[str]):
        pass

    @abstractmethod
    def unpack(self, archive_path: str, dst_path: str):
        pass

    @abstractmethod
    def file_exists(self, path: str) -> bool:
        pass

    @abstractmethod
    def find_temporary_dir_path(self) -> str:
        pass

    @abstractmethod
    def move(self, src: str, dst: str):
        pass




"""
Notifier
========

Integrates external monitoring services for sending infos and alerts
"""

import requests
import json
from time import sleep
from rkd.api.inputoutput import IO
from .model import BackupDefinition


class NotifierInterface(object):
    """
    Notification sending service interface

    <sphinx:notification_types>

    - Starting backup creation
    - Backup was uploaded
    - Failed to upload a backup

    - Starting backup restore
    - Backup was restored
    - Failed to restore a backup

    - Generic: Exception occurred

    </sphinx:notification_types>

    """

    config = {}
    sensitive_data: list
    io: IO

    def __init__(self, config: dict, io: IO):
        self._set_config(config)
        self.sensitive_data = []
        self.io = io

    def _set_config(self, config: dict):
        pass

    def backup_was_uploaded(self, definition: BackupDefinition) -> None:
        pass

    def starting_backup_restore(self, definition: BackupDefinition) -> None:
        pass

    def starting_backup_creation(self, definition: BackupDefinition) -> None:
        pass

    def backup_was_restored(self, definition: BackupDefinition) -> None:
        pass

    def failed_to_upload_backup(self, definition: BackupDefinition, reason) -> None:
        pass

    def failed_to_restore_backup(self, definition: BackupDefinition, reason) -> None:
        pass

    def exception_occurred(self, exception: BaseException):
        pass

    def set_sensitive_data_to_strip_out(self, sensitive_data: list):
        self.sensitive_data = sensitive_data

    def filter_out_sensitive_data(self, input_str: str) -> str:
        for keyword in self.sensitive_data:
            input_str = input_str.replace(keyword, '****')

        return input_str


class MultiplexedNotifiers(object):
    """Proxy that allows to use multiple notifiers at once"""

    _notifiers: list

    def __init__(self, notifiers: dict):
        self._notifiers = list(notifiers.items())

    def __getattr__(self, item):
        return self._create_proxy(item)

    def _create_proxy(self, item):
        def _proxy(*args, **kwargs):
            for handler_name, handler in _proxy.notifiers:
                method = getattr(handler, _proxy.item)
                method(*args, **kwargs)

        _proxy.item = item
        _proxy.notifiers = self._notifiers

        return _proxy


class SlackNotifier(NotifierInterface):
    """Slack/Mattermost notification"""

    _timeout = 300
    _url = ''
    _max_retry_num = 3

    def __init__(self, config: dict, io: IO):
        super().__init__(config, io)
        self._url = config['url']
        self._max_retry_num = int(config.get('max_retry_num', 3))
        self._timeout = int(config.get('connection_timeout', 300))

    def backup_was_uploaded(self, definition: BackupDefinition) -> None:
        self._send(':white_check_mark: Backup was uploaded for ' + str(definition))

    def starting_backup_restore(self, definition: BackupDefinition) -> None:
        self._send(':information_source: Restoring backup of ' + str(definition))

    def starting_backup_creation(self, definition: BackupDefinition) -> None:
        self._send(':information_source: Creating backup of ' + str(definition))

    def backup_was_restored(self, definition: BackupDefinition) -> None:
        self._send(':white_check_mark: Backup was restored for ' + str(definition))

    def failed_to_upload_backup(self, definition: BackupDefinition, reason) -> None:
        self._send(':x: Failed to upload the backup for ' + str(definition) + ', ' + str(reason))

    def failed_to_restore_backup(self, definition: BackupDefinition, reason) -> None:
        self._send(':x: Failed to restore the backup for ' + str(definition) + ', ' + str(reason))

    def exception_occurred(self, exception: BaseException):
        self._send(':bangbang: ' + str(exception))

    def _send(self, msg: str, retry_num: int = 0):
        msg = self.filter_out_sensitive_data(msg)

        self.io.debug('Notifying: %s' % msg)
        self.io.debug('Notifier retry num=%i' % retry_num)

        try:
            response = requests.post(
                self._url, data=json.dumps({'text': msg}),
                headers={'Content-Type': 'application/json'},
                timeout=self._timeout
            )
            if response.status_code != 200:
                raise ValueError(
                    'Request to slack returned an error %s, the response is:\n%s'
                    % (response.status_code, response.text)
                )
        except Exception as e:
            if self._max_retry_num and retry_num < self._max_retry_num:
                self._send(msg, retry_num + 1)
                sleep(1)
                return

            self.io.warn('During sending a notification an unrecoverable error occurred: ' + str(e))


class NotifierFactory(object):
    """Constructs a NotifierInterface implementation basing on given identifier"""

    _SUPPORTED = {
        'slack': SlackNotifier,
        'mattermost': SlackNotifier,
        'none': NotifierInterface,
        'nothing': NotifierInterface,
        '': NotifierInterface,
        'void': NotifierInterface
    }

    @staticmethod
    def create(notifier_type: str, config: dict, io: IO):
        if notifier_type not in NotifierFactory._SUPPORTED:
            raise Exception('Unsupported notifier type "' + notifier_type + '"')

        return NotifierFactory._SUPPORTED[notifier_type](config, io)



from rkd.api.contract import ExecutionContext
from .base import BaseTask


class BackupPreparationTask(BaseTask):
    """Makes a backup and prints to stdout or into a file specified by --target parameter
    """

    def get_name(self) -> str:
        return ':make'

    def get_group_name(self) -> str:
        return ':backup'

    def execute(self, context: ExecutionContext) -> bool:
        if not super().execute(context):
            return False

        return self.call_backup_maker(context, is_backup=True)



"""
Kubernetes - SIDE POD
=====================

Creates a temporary POD that has access to all volumes of original POD.
Temporary POD is attempted to be scheduled closest to the original POD to mitigate the latency
"""
import json
import time
from dataclasses import dataclass
from typing import Tuple, List

from kubernetes.client import ApiException, V1Pod, V1ObjectMeta, V1OwnerReference, V1Scale, V1ScaleSpec
from rkd.api.inputoutput import IO
from .kubernetes_podexec import Transport as KubernetesPodExecTransport


@dataclass
class ReplicaToScale(object):
    kind: str
    name: str
    namespace: str
    replicas: int


class Transport(KubernetesPodExecTransport):
    _image: str
    _timeout: int
    _scale_down: bool

    # dynamic
    _temporary_pod_name: str
    _replicas_to_scale: List[ReplicaToScale]

    def __init__(self, spec: dict, io: IO):
        super().__init__(spec, io)
        self._image = spec.get('image')
        self._timeout = spec.get('timeout')
        self._replicas_to_scale = []
        self._scale_down = bool(spec.get('scaleDown'))

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "oneOf": [
                {"required": ["namespace", "selector"]},
            ],
            "properties": {
                "selector": {
                    "type": "string",
                    "example": "my-label=myvalue",
                    "default": ""
                },
                "namespace": {
                    "type": "string",
                    "example": "prod",
                    "default": "default"
                },
                "image": {
                    "type": "string",
                    "example": "ghcr.io/riotkit-org/backup-maker-env:latest",
                    "default": "ghcr.io/riotkit-org/backup-maker-env:latest"
                },
                "timeout": {
                    "type": "integer",
                    "default": 3600,
                    "example": 3600
                },
                "scaleDown": {
                    "type": "boolean",
                    "default": False,
                    "example": False
                }
            }
        }

    def schedule(self, command: str, definition, is_backup: bool, version: str = "") -> None:
        original_pod_name = self.find_pod_name(self._selector, self._namespace)

        try:
            if self._scale_down:
                self._scale_pod_owner(original_pod_name, self._namespace)

            volumes, volume_mounts = self._copy_volumes_specification_from_existing_pod(
                original_pod_name,
                namespace=self._namespace
            )

            # spawn temporary pod
            self._temporary_pod_name = f"{original_pod_name}-backup"
            self._create_pod(
                pod_name=self._temporary_pod_name,
                specification=self._create_backup_pod_definition(
                    original_pod_name,
                    self._temporary_pod_name,
                    self._timeout,
                    volumes,
                    volume_mounts
                )
            )

            self._execute_in_pod_when_pod_will_be_ready(self._temporary_pod_name, command, definition, is_backup, version)
        except:
            if self._scale_down:
                self._scale_back()

            raise

    def _scale_pod_owner(self, pod_name: str, namespace: str):
        """
        Tell Deployment/ReplicationController/StatefulSet etc. to scale down

        :param pod_name:
        :param namespace:
        :return:
        """

        pod: V1Pod = self._v1_core_api.read_namespaced_pod(name=pod_name, namespace=namespace)
        metadata: V1ObjectMeta = pod.metadata
        owners: List[V1OwnerReference] = metadata.owner_references

        if not owners:
            self.io().warn("No POD owner found through owner references")
            return

        self._scale_by_owner_references(owners, namespace)

    def _scale_by_owner_references(self, owners: List[V1OwnerReference], namespace: str):
        for owner in owners:
            if owner.kind == "ReplicaSet":
                rs = self.v1_apps_api.read_namespaced_replica_set(name=owner.name, namespace=namespace)
                metadata: V1ObjectMeta = rs.metadata
                rs_owners: List[V1OwnerReference] = metadata.owner_references

                self._scale_by_owner_references(rs_owners, namespace)
                continue

            elif owner.kind != "Deployment":
                self.io().warn(f"Unsupported controller type '{owner.kind}', will not attempt to scale it")
                continue

            deployment_as_dict = json.loads(self.v1_apps_api.read_namespaced_deployment(name=owner.name,
                                                                                        namespace=namespace,
                                                                                        _preload_content=False).data)

            self._replicas_to_scale.append(ReplicaToScale(
                kind='Deployment',
                name=owner.name,
                namespace=namespace,
                replicas=deployment_as_dict['spec'].get('replicas', 0)
            ))
            self._scale(owner.name, namespace, 0)

    def _scale(self, name: str, namespace: str, replicas: int):
        """
        Scale down given Deployment/ReplicationController/StatefulSet

        :param name:
        :param namespace:
        :param replicas:
        :return:
        """

        self.io().info(f"Scaling deployment/{name} in {namespace} namespace to replicas '{replicas}'")
        scale = V1Scale(
            metadata=V1ObjectMeta(name=name, namespace=namespace),
            spec=V1ScaleSpec(
                replicas=replicas
            )
        )
        self.v1_apps_api.replace_namespaced_deployment_scale(name, namespace, scale)

        # then wait for it to be applied
        for i in range(0, 3600):
            deployment_as_dict = json.loads(self.v1_apps_api.read_namespaced_deployment(name=name,
                                                                                        namespace=namespace,
                                                                                        _preload_content=False).data)

            current_replicas_num = int(deployment_as_dict['spec']['replicas'])

            if current_replicas_num == int(replicas):
                self.io().info(f"POD's controller scaled to {replicas}")
                break

            self.io().debug(f"Waiting for cluster to scale POD's controller to {replicas},"
                            f" currently: {current_replicas_num}")
            time.sleep(1)

    def _scale_back(self):
        """
        Bring back service pods

        :return:
        """

        for kind_object in self._replicas_to_scale:
            self._scale(kind_object.name, kind_object.namespace, kind_object.replicas)

    def _create_pod(self, pod_name: str, specification: dict):
        self.io().info(f"Creating temporary POD '{pod_name}'")

        try:
            self._v1_core_api.create_namespaced_pod(namespace=self._namespace, body=specification)

        except ApiException as e:
            if e.reason == "Conflict" and "AlreadyExists" in str(e.body):
                raise Exception(f"POD '{pod_name}' already exists or is terminating, "
                                f"please wait a moment - cannot start process in parallel, "
                                f"it may break something") from e

            raise e

    def _terminate_pod(self, pod_name: str):
        self.io().info("Clean up - deleting temporary POD")
        self._v1_core_api.delete_namespaced_pod(namespace=self._namespace, name=pod_name)

    def _copy_volumes_specification_from_existing_pod(self, pod_name: str, namespace: str) -> Tuple[dict, list]:
        self.io().debug(f"Copying volumes specification from source pod={pod_name}")
        self.wait_for_pod_to_be_ready(pod_name, namespace)
        pod = json.loads(self._v1_core_api.read_namespaced_pod(pod_name, namespace, _preload_content=False).data)

        try:
            pod_volumes = pod['spec']['volumes']

        except KeyError:
            return {}, []

        mounts = []
        _already_added_mounts = []

        # grab mounts from all containers, avoid duplications
        for container in pod['spec']['containers']:
            self.io().debug(f"Processing container with image={container['image']}")

            try:
                volume_mounts = container['volumeMounts']
            except KeyError:
                self.io().debug(f"Container with image '{container['image']}' does not have volumeMounts")
                continue

            for mount in volume_mounts:
                target_path = mount['mountPath']

                # do not mount twice :-)
                if target_path in _already_added_mounts:
                    self.io().warn(f"Container with image={container['image']} has overlapping mount "
                                   f"of other container")
                    continue

                mounts.append(mount)

        return pod_volumes, mounts

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        """
        After operation try to clean up - terminate the temporary POD

        :param exc_type:
        :param exc_val:
        :param exc_t:
        :return:
        """

        try:
            if hasattr(self, '_temporary_pod_name'):
                self._terminate_pod(self._temporary_pod_name)
        finally:
            self._scale_back()

    def _create_backup_pod_definition(self, original_pod_name: str, backup_pod_name: str, timeout: int,
                                      volumes: dict, volume_mounts: list) -> dict:
        return {
            'apiVersion': 'v1',
            'kind': 'Pod',
            'metadata': {
                "name": backup_pod_name,
                "namespace": self._namespace,
                "labels": {
                    "riotkit.org/original-pod": original_pod_name,
                }
            },
            'spec': {
                'containers': [
                    {
                        'image': self._image,
                        'name': backup_pod_name,
                        'command': ["/bin/sh"],
                        'args': ['-c', f'sleep {str(timeout)}'],
                        'restartPolicy': 'never',
                        'volumeMounts': volume_mounts
                    }
                ],
                'volumes': volumes
            }
        }



from dataclasses import dataclass
from typing import List
from abc import ABC, abstractmethod
from jsonschema import validate, draft7_format_checker, ValidationError

from .bin import RequiredBinary
from .exception import SpecificationError
from .transports.base import TransportInterface
from .schema import create_example_from_attributes


@dataclass
class ServerAccess(object):
    """
    Backup Repository user access
    """

    url: str
    token: str

    @staticmethod
    def from_config(config: dict):
        return ServerAccess(config['url'], config['token'])

    def get_url(self):
        return self.url

    def get_token(self):
        return self.token

    def build_url(self, endpoint: str) -> str:
        url = self.url.rstrip('/') + '/' + endpoint.lstrip('/')
        qs = '?'

        return url + qs.replace('?&', '?')


class Encryption(object):
    """
    Cryptography support (using OpenSSL command implementation)
    """

    _name: str
    _passphrase: str
    _public_key_path: str
    _private_key_path: str
    _user_email: str

    def __init__(self, name: str, passphrase: str, email: str, public_key_path: str, private_key_path: str):

        self._name = name
        self._passphrase = passphrase
        self._public_key_path = public_key_path
        self._private_key_path = private_key_path
        self._user_email = email

    @staticmethod
    def from_config(name: str, config: dict):
        is_encrypting = config.get('method', '') != ''

        if is_encrypting:
            if "private_key_path" not in config and "public_key_path" not in config:
                raise KeyError("private_key_path|or|public_key_path")

        return Encryption(
            name=name,
            passphrase=config.get('passphrase', ''),
            email=config['email'],
            public_key_path=config['public_key_path'],
            private_key_path=config['private_key_path']
        )

    def recipient(self):
        return self._user_email

    def get_passphrase(self) -> str:
        return self._passphrase

    def name(self) -> str:
        return self._name

    def get_public_key_path(self) -> str:
        if not self._public_key_path:
            return self.get_private_key_path()

        return self._public_key_path

    def get_private_key_path(self) -> str:
        return self._private_key_path


class BackupDefinition(ABC):
    """
    Configuration definition containing credentials, tokens, ids
    """

    _access: ServerAccess
    _encryption: Encryption
    _collection_id: str
    _name: str
    _spec: dict
    _transport: TransportInterface

    def __init__(self, access: ServerAccess, collection_id: str, encryption: Encryption,
                 name: str, spec: dict, transport: TransportInterface):
        self._access = access
        self._encryption = encryption
        self._collection_id = collection_id
        self._name = name
        self._spec = spec
        self._transport = transport

    @staticmethod
    def from_config(cls, config: dict, name: str):
        cls.validate_spec(config['spec'])

        return cls(
            access=config['meta']['access'],
            collection_id=config['meta']['collection_id'],
            encryption=config['meta']['encryption'],
            name=name,
            spec=config['spec'],
            transport=config['meta']['transport']
        )

    @staticmethod
    def get_spec_defaults() -> dict:
        """Specification default values"""

        return {}

    @classmethod
    def validate_spec(cls, spec: dict):
        spec_schema = cls.get_specification_schema()

        if spec_schema:
            try:
                validate(instance=spec, schema=spec_schema, format_checker=draft7_format_checker)
            except ValidationError as exc:
                raise SpecificationError(str(exc))

    @staticmethod
    @abstractmethod
    def get_specification_schema() -> dict:
        """A JSON-schema of 'spec' element of the configuration.
        'spec' is the specific configuration for given type of definition, for example MySQL will have a spec
        with instructions how to reach the database
        """
        pass

    @staticmethod
    def join_paths_quotes_into_string(paths: list):
        joined = ""

        for path in paths:
            joined += '"' + path + '" '

        return joined

    def access(self) -> ServerAccess:
        return self._access

    def encryption(self) -> Encryption:
        return self._encryption

    def get_collection_id(self) -> str:
        return self._collection_id

    def transport(self, binaries: List[RequiredBinary]) -> TransportInterface:
        self._transport.prepare_environment(binaries)
        return self._transport

    def get_transport_required_tools(self) -> List[RequiredBinary]:
        return self._transport.get_required_binaries()

    def get_sensitive_information(self) -> list:
        return []

    def name(self) -> str:
        return self._name

    def __repr__(self):
        return 'Definition<name=' + self._name + ',collection_id=' + str(self.get_collection_id()) + '>'

    @classmethod
    def get_example_configuration(cls) -> dict:
        schema = cls.get_specification_schema()

        if not schema or 'properties' not in schema:
            return {}

        return {
            'meta': {
                'type': cls.__module__,
                'access': 'my_backup_server',
                'encryption': 'enc_backup_1',
                'collection_id': '61792136-94d5-4670-9c69-950257467c56',
                'transport': 'local'
            },
            'spec': create_example_from_attributes(schema['properties'])
        }



from abc import ABC as AbstractClass, abstractmethod
from typing import List

from ..bin import RequiredBinary
from ..model import BackupDefinition


class AdapterInterface(AbstractClass):
    @staticmethod
    @abstractmethod
    def create_definition(config: dict, name: str) -> BackupDefinition:
        """
        Creates a configuration object that will contain instructions how to e.g. access the database, where to find
        files etc.
        """

        pass

    @abstractmethod
    def create_backup_instruction(self, definition: BackupDefinition) -> str:
        pass

    @abstractmethod
    def create_restore_instruction(self, definition: BackupDefinition) -> str:
        pass

    @abstractmethod
    def get_required_binaries(self) -> List[RequiredBinary]:
        """
        Lists required binaries
        :return:
        """

        return []



import os
import sys
from abc import abstractmethod
from subprocess import Popen, PIPE
from time import sleep
from typing import List, Union, Optional
from jsonschema import validate, draft7_format_checker, ValidationError
from rkd.api.inputoutput import IO

from ..bin import RequiredBinary
from ..exception import SpecificationError
from ..inputoutput import StreamableBuffer
from ..schema import create_example_from_attributes


def create_backup_maker_command(command: str, definition, is_backup: bool,
                                version: str = "", prepend: list = None, bin_path: str = '') -> List[str]:
    args = [
        "/usr/bin/env"
    ]

    if bin_path:
        args += [
            bin_path + "/tracexit", f"env:PATH+BIN={bin_path}"
        ]

    if prepend:
        args += prepend

    args += [
        "br-backup-maker",
        "make" if is_backup else "restore",
        "--url", definition.access().url,
        "--collection-id", definition.get_collection_id(),
        "--auth-token", definition.access().token,
        "-c", command,
        "--recipient", definition.encryption().recipient(),
        "--log-level", "info"
    ]

    if is_backup:
        if os.path.realpath(definition.encryption().get_public_key_path()):
            args += ["--key", "/tmp/.gpg.pub"]
    else:
        if os.path.realpath(definition.encryption().get_private_key_path()):
            args += ["--private-key", "/tmp/.gpg.key"]

    if definition.encryption().get_passphrase():
        args += ["--passphrase", definition.encryption().get_passphrase()]

    if not is_backup:
        args += ["--version", version]

    return args


class TransportInterface(object):
    """
    Transport
    =========

    Defines how to trigger adapters. It is not that simple - there are many possibilities:
        - Simply just execute on local shell
        - Execute inside running Docker container
        - Execute in Kubernetes container
        - Execute on remote machine
    """

    _spec: dict
    _io: IO

    def __init__(self, spec: dict, io: IO):
        self._spec = spec
        self._io = io

    def __enter__(self) -> 'TransportInterface':
        """
        Start using the transport. Here could be placed a code that will eg. spawn a docker container
        or notify the user - any action before usage of the Transport
        :return:
        """

        return self

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        """
        Finalize usage of the Transport - close what remained, clean up
        :return:
        """

        pass

    @staticmethod
    @abstractmethod
    def get_specification_schema() -> dict:
        """
        Should return a DICT with JSON Schema defining the "spec" of a transport in YAML

        :return:
        """

        return {}

    @classmethod
    def get_example_configuration(cls) -> dict:
        schema = cls.get_specification_schema()

        if not schema or 'properties' not in schema:
            return {}

        return {
            'type': cls.__module__,
            'spec': create_example_from_attributes(schema['properties'])
        }

    @classmethod
    def validate_spec(cls, spec: dict):
        spec_schema = cls.get_specification_schema()

        if spec_schema:
            try:
                validate(instance=spec, schema=spec_schema, format_checker=draft7_format_checker)
            except ValidationError as exc:
                raise SpecificationError('Linked transport "spec" section parsing error, ' + str(exc))

    @abstractmethod
    def prepare_environment(self, binaries: List[RequiredBinary]) -> None:
        pass

    @abstractmethod
    def schedule(self, command: str, definition, is_backup: bool, version: str = "") -> None:
        """
        Schedule a backup
        """
        pass

    @abstractmethod
    def watch(self) -> bool:
        """
        Watches output of the scheduled command and returns boolean to indicate result
        """

        pass

    def get_failure_details(self) -> str:
        """
        Optionally raise a specific exception with more details
        """

        return ''

    def io(self) -> IO:
        return self._io

    def _exec_command(self, command: Union[str, List[str]], stdin: Optional[StreamableBuffer] = None,
                      env: dict = None) -> StreamableBuffer:

        proc_env = dict(os.environ)

        if env:
            proc_env.update(env)

        self.io().debug('_exec_command({command})'.format(command=command))
        proc = Popen(command, shell=type(command) == str,
                     stdout=PIPE,
                     stderr=sys.stderr.fileno(),
                     env=env,
                     text=True)

        def close_stream():
            proc.stdout.close()
            proc.terminate()
            sleep(1)

        return StreamableBuffer(
            io=self._io,
            read_callback=proc.stdout.read,
            close_callback=close_stream,
            eof_callback=lambda: proc.poll() is not None,
            is_success_callback=lambda: proc.poll() == 0,
            has_exited_with_failure=lambda: proc.poll() is not None and proc.poll() >= 1,
            description='command <{}>'.format(command),
            buffer=proc.stdout,
            parent=stdin
        )

    def get_required_binaries(self):
        return []



"""
Filesystem Adapter
==================

Packs files and directories into TAR.GZ packages
"""
import os
from typing import List

from .base import AdapterInterface
from ..bin import RequiredBinary
from ..model import BackupDefinition


class Definition(BackupDefinition):
    """
    Configuration
    """

    @staticmethod
    def get_spec_defaults() -> dict:
        return {}

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["paths"],
            "properties": {
                "paths": {
                    "type": "array",
                    "items": {
                        "type": "string"
                    },
                    "example": ["/var/lib/jenkins"]
                }
            }
        }

    def get_backup_parameters(self):
        """
        Inside a package there could be multiple directories and files packaged from multiple paths
        Every path is quoted and checked before usage
        :return:
        """

        paths = []

        for path in self._spec.get('paths'):
            normalized = os.path.abspath(os.path.expanduser(path))

            if not normalized:
                continue

            paths.append('"{}"'.format(normalized))

        parameters = '-cz -f - ' + ' '.join(paths)

        return parameters

    def get_restore_parameters(self):
        """
        We unpack always at the root directory, because we allow to pack multiple paths
        :return:
        """

        return '-xzf - -C /'


class Adapter(AdapterInterface):
    """
    Defines how to make backup of files and directories
    """

    def create_backup_instruction(self, definition: Definition) -> str:
        """
        Pack files into a TAR.GZ
        """

        return 'tar %s | cat' % definition.get_backup_parameters()

    def create_restore_instruction(self, definition: Definition) -> str:
        """
        Unpack files from the TAR.GZ
        """

        return 'tar %s' % definition.get_restore_parameters()

    @staticmethod
    def create_definition(config: dict, name: str) -> Definition:
        return BackupDefinition.from_config(Definition, config, name)

    def get_required_binaries(self) -> List[RequiredBinary]:
        return []



"""
Helpers for testing in PyTest
"""


from bahub.model import ServerAccess, Encryption
from bahub.adapters.filesystem import Definition as FilesystemBackupDefinition


def run_transport(definition, transport) -> bool:
    """
    Runs schedule & watch for a transport

    :param definition:
    :param transport:
    :return:
    """

    with definition.transport(binaries=[]):
        transport.schedule(
            command="--mocked--", definition=definition,
            is_backup=True, version=""
        )

        return transport.watch()


def create_example_fs_definition(transport) -> FilesystemBackupDefinition:
    """
    Creates an example Filesystem adapter definition for given Transport

    :param transport:
    :return:
    """

    return FilesystemBackupDefinition.from_config(cls=FilesystemBackupDefinition, config={
        "meta": {
            "access": ServerAccess(
                url="http://localhost:8080",
                token="test"
            ),
            "collection_id": "1111-2222-3333-4444",
            "encryption": Encryption.from_config(name="enc", config={
                "passphrase": "riotkit",
                "email": "test@riotkit.org",
                "public_key_path": "",
                "private_key_path": "test/env/config_factory_test/gpg-key.asc"
            }),
            "transport": transport
        },
        "spec": {
            "paths": ["/app"]
        },
    }, name="fs")



"""
Temporary Docker Transport (bahub.transports.temporarydocker)
=============================================================

Places a temporary container to copy data from other container while the second container will be shutted down for
maintenance time
"""
import subprocess

from rkd.api.inputoutput import IO
from docker.models.containers import Container
from .docker import Transport as RegularDockerTransport, DockerFilesystemTransport


class Transport(RegularDockerTransport):
    """
    Behaves exactly the same as regular Docker Transport, but on a temporary container
    Even the methods are used from parent Docker Transport - and the `self._container_name` is replaced
    with a temporary container name
    """

    _temp_image: str
    _spec: dict
    _should_stop_original: bool
    _should_pull_image: bool

    original_container: Container  # original container
    container: Container           # temporary container

    def __init__(self, spec: dict, io: IO):
        self._spec = spec
        super().__init__(spec, io)
        self._temp_image = spec.get('temp_container_image')
        self._shell = spec.get('shell', '/bin/bash')
        self._should_stop_original = spec.get('stop', True)
        self._should_pull_image = spec.get('pull', True)

    def _populate_container_information(self):
        self._container_name = self._spec.get('orig_container')
        self.original_container = self._client.containers.get(self._container_name)

    @staticmethod
    def get_specification_schema() -> dict:
        return {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "type": "object",
            "required": ["temp_container_image", "orig_container"],
            "properties": {
                "orig_container": {
                    "type": "string",
                    "example": "bahub_adapter_integrations_db_mysql_1"
                },
                "temp_container_image": {
                    "type": "string",
                    "example": "mariadb:10.3"
                },
                "shell": {
                    "type": "string",
                    "example": "/bin/bash",
                    "default": "/bin/bash"
                },
                "stop": {
                    "type": "boolean",
                    "example": True,
                    "default": True
                }
            }
        }

    def __enter__(self) -> 'Transport':
        """
        Spawns a temporary container
        :return:
        """

        is_there_original_container = hasattr(self, "original_container")

        # stop the original container at first
        if is_there_original_container and self._should_stop_original:
            self.io().info('Stopping original container {}'.format(self.original_container.id))
            self.original_container.stop()

        try:
            self.io().info('Creating a temporary container...')
            host_config = self.client.api.create_host_config(volumes_from=[self.original_container.id])

            if self._should_pull_image:
                subprocess.check_call(['docker', 'pull', self._temp_image])

            info = self.client.api.create_container(
                image=self._temp_image,
                entrypoint=['sleep'],
                command=[str(86400 * 5)],
                host_config=host_config
            )
            self._container = self.client.containers.get(info['Id'])
            self.io().debug('Temporary container was spawned')

            self._container.start()
            self.io().debug('Temporary container was started')

            # will allow injection of required binaries into container
            self.fs = DockerFilesystemTransport(self._container)
        except:
            if is_there_original_container and self._should_stop_original:
                self.io().error("Error while starting temporary container, restoring original container")
                self.original_container.start()

            raise

        return self

    def __exit__(self, exc_type, exc_val, exc_t) -> None:
        try:
            self.io().info('Killing temporary container')
            self.container.kill()
        finally:
            self.io().debug('Removing temporary container')
            self.container.remove(force=True)

            if self._should_stop_original:
                self.io().info('Bringing back the orginal container')
                self.original_container.start()



"""
Generic Kubernetes methods for building Kubernetes transports
"""


import os
from typing import List, Callable
import yaml
from tempfile import TemporaryDirectory
from kubernetes.stream.ws_client import WSClient, ERROR_CHANNEL
from rkd.api.inputoutput import IO
from kubernetes import client
from kubernetes.stream import stream
from ..fs import FilesystemInterface


class ExecResult(object):
    """
    Result of operation like `kubectl exec`
    """

    _process: WSClient
    _io: IO

    def __init__(self, process: WSClient, io: IO):
        self._process = process
        self._io = io

    def read(self) -> str:
        """
        Wait till process exit, then read output
        """

        self._process.run_forever()
        return self._process.read_all()

    def watch(self, printer: Callable) -> None:
        """
        Watches process for output
        """

        while self._process.is_open():
            self._process.update(timeout=1)

            out = [
                self._process.readline_stdout() if self._process.peek_stdout() else "",
                self._process.readline_stderr() if self._process.peek_stderr() else ""
            ]

            for line in out:
                if line:
                    printer(line)

    def is_still_running(self) -> bool:
        return self._process.is_open()

    def has_exited_with_success(self) -> bool:
        if self.is_still_running():
            return True

        errors = yaml.load(self._process.read_channel(ERROR_CHANNEL), yaml.FullLoader)

        if errors and "details" in errors:
            for error in errors['details']['causes']:
                if "reason" not in error:
                    self._io.error(error['message'])
                    return False

                if error['reason'] == 'ExitCode' and int(error['message']) > 0:
                    self._io.error(f"Process inside POD exited with status {int(error['message'])}")
                    return False

        return True


def pod_exec(pod_name: str, namespace: str, cmd: List[str], io: IO) -> ExecResult:
    """
    Execute a command inside a POD
    """

    return ExecResult(
        stream(
            client.CoreV1Api().connect_get_namespaced_pod_exec,
            pod_name,
            namespace,
            command=cmd,
            stderr=True,
            stdout=True,
            stdin=False,
            _preload_content=False
        ),
        io
    )


class KubernetesPodFilesystem(FilesystemInterface):
    io: IO
    pod_name: str
    namespace: str

    def __init__(self, pod_name: str, namespace: str, io: IO):
        self.io = io
        self.pod_name = pod_name
        self.namespace = namespace

    def _exec(self, cmd: List[str], msg: str, exit_code_hack: bool = False):
        if exit_code_hack:
            cmd = ["/bin/sh", "-c", (" ".join(cmd)) + " && echo '@<br-exit-ok>'"]

        proc = pod_exec(self.pod_name, self.namespace, cmd, self.io)
        result = proc.read()

        assert proc.has_exited_with_success(), f"{msg}. {result}"

        if exit_code_hack:
            assert "@<br-exit-ok>" in result, f"Process exited with failure. Output: {result}"

    def force_mkdir(self, path: str):
        self._exec(["mkdir", "-p", path], "mkdir inside POD failed, cannot create directory", exit_code_hack=True)

    def download(self, url: str, destination_path: str):
        self._exec(
            ["curl", "-s", "-L", "--output", destination_path, url],
            f"curl inside POD failed, cannot download file from '{url}' to '{destination_path}' path inside POD",
            exit_code_hack=True
        )

    def delete_file(self, path: str):
        try:
            self._exec(["rm", path], f"Cannot remove file inside POD at path '{path}' (inside POD)",
                       exit_code_hack=True)

        except AssertionError:
            self.io.debug(f"Cannot remove file inside POD at path '{path}' (inside POD). Maybe file does not exist")
            pass

    def link(self, src: str, dst: str):
        self._exec(["ln", "-s", src, dst], f"Cannot make symbolic link from '{src}' to '{dst}' (inside POD)",
                   exit_code_hack=True)

    def make_executable(self, path: str):
        self._exec(["chmod", "+x", path], f"Cannot make file executable at path '{path}' (inside POD)",
                   exit_code_hack=True)

    def copy_to(self, local_path: str, dst_path: str):
        process = stream(
            client.CoreV1Api().connect_get_namespaced_pod_exec,
            self.pod_name,
            self.namespace,
            command=["/bin/sh", "-c", f"cat - > {dst_path}"],
            stderr=True,
            stdout=True,
            stdin=True,
            _preload_content=False
        )

        with open(local_path, 'rb') as f:
            while process.is_open():
                process.update(timeout=1)
                process.write_stdin(f.read(1024*1024))

                if f.tell() == os.fstat(f.fileno()).st_size:
                    process.close()

    def pack(self, archive_path: str, src_path: str, files_list: List[str]):
        if not files_list:
            files_list = ["*", ".*"]

        self._exec(
            ["tracexit", f"env:PWD={src_path}", "tar", "-zcf", archive_path] + files_list,
            f"Cannot pack files from {src_path} into {archive_path} (inside POD)"
        )

    def unpack(self, archive_path: str, dst_path: str):
        self._exec(
            ["tar", "xf", archive_path, "--directory", dst_path],
            f"Cannot unpack files from '{archive_path}' to '{dst_path}'"
        )

    def file_exists(self, path: str) -> bool:
        try:
            self._exec(["test", "-f", path], f"File does not exist", exit_code_hack=True)

        except AssertionError:
            return False

        return True

    def find_temporary_dir_path(self) -> str:
        return TemporaryDirectory().name

    def move(self, src: str, dst: str):
        self._exec(
            ["mv", src, dst],
            f"Cannot move file {src} to {dst} inside POD",
            exit_code_hack=True
        )



from .docker import Transport as DockerTransport
from .sh import Transport as ShellTransport
from .docker_sidecontainer import Transport as TemporaryDockerTransport
from .kubernetes_podexec import Transport as KubernetesPodExecTransport
from .kubernetes_sidepod import Transport as KubernetesSidePodTransport


def transports():
    return [ShellTransport, DockerTransport, TemporaryDockerTransport,
            KubernetesPodExecTransport, KubernetesSidePodTransport]
